%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [dy,fluxy] = odes_pa_sa_biofilm(t, y, INFO)

% t

nmodel = INFO.nmodel;
N = INFO.N;
ns = INFO.ns;
param = INFO.param;
condit = INFO.condit;

% Set conditions

L = condit(1);
glcdiffa = 3600*(1e6)^2/(100)^2*condit(2);    % Aqueous glucose diffusion coefficient, micron^2/h
o2diffa = 3600*(1e6)^2/(100)^2*condit(3);     % Aqueous oxygen diffusion coefficient, micron^2/h
acediffa = 3600*(1e6)^2/(100)^2*condit(4);    % Aqueous acetate diffusion coefficient, micron^2/h
succdiffa = 3600*(1e6)^2/(100)^2*condit(5);   % Aqueous succinate diffusion coefficient, micron^2/h
lacdiffa = 3600*(1e6)^2/(100)^2*condit(6);    % Aqueous lactate diffusion coefficient, micron^2/h
pyodiffa = 3600*(1e6)^2/(100)^2*condit(7);    % Aqueous pyocyanin diffusion coefficient, micron^2/h
celldiffa = 3600*(1e6)^2/(100)^2*condit(25);    % Aqueous pyocyanin diffusion coefficient, micron^2/h
kxt = 3600*(1e6)^2/(100)^2*condit(8);                 % PA aerotaxis coefficient, (mmol PA)(micron)/(mmol O2)/h
glcmt = 3600*1e6/100*condit(9);               % Glucose mass transfer coefficient, micron/h
o2mt = 3600*1e6/100*condit(10);               % Oxygen mass transfer coefficient, micron/h
acemt = 3600*1e6/100*condit(11);              % Acetate mass transfer coefficient, micron/h
succmt = 3600*1e6/100*condit(12);             % Succinate mass transfer coefficient, micron/h
lacmt = 3600*1e6/100*condit(13);              % Lactate mass transfer coefficient, micron/h
pyomt = 3600*1e6/100*condit(14);              % Pyocyanin mass transfer coefficient, micron/h
xmax = condit(15);
N = condit(16);
ns = condit(17);
zs = condit(18);
glcb = condit(19);
o2L = condit(20);
aceb = condit(21);
succb = condit(22);
lacb = condit(23);
pyob = condit(24);
kpyod = condit(26);
kpadm = condit(27);
ksadm = condit(28);
td = condit(29);

% Set extracellular state variables

for i=1:N+1
    xbm(i) = y(1+(i-1)*ns);     % PA biomass concentration, g/L
    zbm(i) = y(2+(i-1)*ns);     % SA biomass concentration, g/L
    glc(i) = y(3+(i-1)*ns);     % Glucose concentration, mmol/L
    o2(i) = y(4+(i-1)*ns);      % Oxygen concentration, mmol/L
    ace(i) = y(5+(i-1)*ns);     % Acetate concentration, mol/L
    succ(i) = y(6+(i-1)*ns);    % Succinate concentration, mol/L
    lac(i) = y(7+(i-1)*ns);     % Lactate concentration, mol/L
    pyo(i) = y(8+(i-1)*ns);     % Inhibitor concentration, mol/L    
end

% The elements of the flux matrix have the sign given to them by the
% coefficients in the Cost vector in main. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% flux(k,i) = 3*v_144 + v_832 - v_931 
% The penalty is an array containing the feasibility LP problem optimal
% objective function value for each model. 

% Update bounds and solve for fluxes

[flux,penalty] = solveModel(t,y,INFO);
fluxy = [];
for i=1:N+1
    
% Set exchange fluxes
    
    xmu(i) = flux(i,1);           % PA growth rate, h-1
    xvace(i) = flux(i,2);         % PA acetate secretion rate, mmol/g/h
    xvsucc(i) = flux(i,3);        % PA succinate secretion rate, mmol/g/h
    xvglc(i) = flux(i,4);         % PA glucose uptake rate, mmol/g/h
    xvo2(i) = flux(i,5);          % PA oxygen uptake rate, mmol/g/h
    xvlac(i) = flux(i,6);         % PA lactate uptake rate, mmol/g/h
    xvpyo(i) = flux(i,7);         % PA pyocyanin secretion rate, mmol/g/h
    kpad(i) = 0;
        
    if y(end-nmodel+i) > 1e-6
        xmu(i) = 0;
        xvace(i) = 0;
        xvsucc(i) = 0; 
        penalty(i) = 0;
        kpad(i) = kpadm*(1-exp(-(t-INFO.INFtimes(i))/td));
    end
    
    zmu(i) = flux(N+1+i,1);           % PA growth rate, h-1
    zvace(i) = flux(N+1+i,2);         % PA acetate secretion rate, mmol/g/h
    zvlac(i) = flux(N+1+i,3);         % PA succinate secretion rate, mmol/g/h
    zvglc(i) = flux(N+1+i,4);         % PA glucose uptake rate, mmol/g/h
    zvo2(i) = flux(N+1+i,5);          % PA oxygen uptake rate, mmol/g/h
    zvsucc(i) = flux(N+1+i,6);        % PA lactate uptake rate, mmol/g/h
    ksad(i) = 0;
    
    if y(end-nmodel+N+1+i) > 1e-6
        zmu(i) = 0;
        zvace(i) = 0;
        zvlac(i) = 0;
        penalty(N+1+i) = 0;
        ksad(i) = ksadm*(1-exp(-(t-INFO.INFtimes(N+1+i))/td));
        
    end
   
% Calculate biofilm diffusion coefficients

    xtot(i) = xbm(i)+zbm(i);
    glcdiff(i) = glcdiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    o2diff(i) = o2diffa*(1-0.19*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    acediff(i) = acediffa*(1-0.36*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    succdiff(i) = succdiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    lacdiff(i) = lacdiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    pyodiff(i) = pyodiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    celldiff(i) = celldiffa*(1-0.33*xtot(i)^0.92/(11.19+0.27*xtot(i)^0.99));
    fluxy = [fluxy xmu(i) xvace(i) xvlac(i) xvglc(i) xvo2(i) xvsucc(i) xvpyo(i) zmu(i) zvace(i) zvlac(i) zvglc(i) zvo2(i) zvsucc(i) kpad(i) ksad(i)];
end

% Set lower boundary conditions
        
xbm0 = xbm(2);
zbm0 = zbm(2);
glc0 = glc(2)+2*glcmt*zs*(glcb-glc(1))/glcdiff(1);
o20 = o2(2)+2*o2mt*zs*(0-o2(1))/o2diff(1);
ace0 = ace(2)+2*acemt*zs*(aceb-ace(1))/acediff(1);
succ0 = succ(2)+2*succmt*zs*(succb-succ(1))/succdiff(1);
lac0 = lac(2)+2*lacmt*zs*(lacb-lac(1))/lacdiff(1);
pyo0 = pyo(2)+2*pyomt*zs*(pyob-pyo(1))/pyodiff(1);
        
% Set upper boundary conditions
        
xbmf = xbm(N); 
zbmf = zbm(N);
glcf = glc(N);
o2f = o2(N)+2*o2mt*zs*(o2L-o2(N+1))/o2diff(N+1);
acef = ace(N);
succf = succ(N);
lacf = lac(N);
pyof = pyo(N);

% Evaluate equations

dy = [];

for i=1:N+1
    
% Compute second order derivatives

    if i==1
        xbmd2(i) = (xbm(i+1)-2*xbm(i)+xbm0)/zs^2;
        zbmd2(i) = (zbm(i+1)-2*zbm(i)+zbm0)/zs^2;
        glcd2(i) = (glc(i+1)-2*glc(i)+glc0)/zs^2;
        o2d2(i) = (o2(i+1)-2*o2(i)+o20)/zs^2;
        aced2(i) = (ace(i+1)-2*ace(i)+ace0)/zs^2;
        succd2(i) = (succ(i+1)-2*succ(i)+succ0)/zs^2;
        lacd2(i) = (lac(i+1)-2*lac(i)+lac0)/zs^2;
        pyod2(i) = (pyo(i+1)-2*pyo(i)+pyo0)/zs^2;
        xbmd1(i) = (xbm(i+1)-xbm0)/(2*zs);
        o2d1(i) = (o2(i+1)-o20)/(2*zs);
                 
    elseif i==N+1
        xbmd2(i) = (xbmf-2*xbm(i)+xbm(i-1))/zs^2;
        zbmd2(i) = (zbmf-2*zbm(i)+zbm(i-1))/zs^2;
        glcd2(i) = (glcf-2*glc(i)+glc(i-1))/zs^2;
        o2d2(i) = (o2f-2*o2(i)+o2(i-1))/zs^2;
        aced2(i) = (acef-2*ace(i)+ace(i-1))/zs^2;
        succd2(i) = (succf-2*succ(i)+succ(i-1))/zs^2;
        lacd2(i) = (lacf-2*lac(i)+lac(i-1))/zs^2;
        pyod2(i) = (pyof-2*pyo(i)+pyo(i-1))/zs^2;
        xbmd1(i) = (xbmf-xbm(i-1))/(2*zs);
        o2d1(i) = (o2f-o2(i-1))/(2*zs);
        
    else
        xbmd2(i) = (xbm(i+1)-2*xbm(i)+xbm(i-1))/zs^2;
        zbmd2(i) = (zbm(i+1)-2*zbm(i)+zbm(i-1))/zs^2;
        glcd2(i) = (glc(i+1)-2*glc(i)+glc(i-1))/zs^2;
        o2d2(i) = (o2(i+1)-2*o2(i)+o2(i-1))/zs^2;
        aced2(i) = (ace(i+1)-2*ace(i)+ace(i-1))/zs^2;
        succd2(i) = (succ(i+1)-2*succ(i)+succ(i-1))/zs^2;
        lacd2(i) = (lac(i+1)-2*lac(i)+lac(i-1))/zs^2;
        pyod2(i) = (pyo(i+1)-2*pyo(i)+pyo(i-1))/zs^2;
        xbmd1(i) = (xbm(i+1)-xbm(i-1))/(2*zs);
        o2d1(i) = (o2(i+1)-o2(i-1))/(2*zs);
       
    end

    % Evaluate RHS of ODEs
    
    dxbm(i) = xmu(i)*xbm(i)*(1-xtot(i)/xmax)-kpad(i)*xbm(i);%+kxt*(xbm(i)*o2d2(i)+xbmd1(i)*o2d1(i))*(1-xtot(i)/xmax); %-(kxt/o2(i))*(xbm(i)*o2d2(i)+xbmd1(i)*o2d1(i)-((xbm(i)/o2(i))*o2d1(i)*o2d1(i)))*(1-xtot(i)/xmax); %+celldiff(i)*xbmd2(i);
    dzbm(i) = zmu(i)*zbm(i)*(1-xtot(i)/xmax)-kpyod*pyo(i)*zbm(i)-ksad(i)*zbm(i);%+celldiff(i)*zbmd2(i);
    dglc(i) = xvglc(i)*xbm(i)+zvglc(i)*zbm(i)+glcdiff(i)*glcd2(i);
    do2(i) = xvo2(i)*xbm(i)+zvo2(i)*zbm(i)+o2diff(i)*o2d2(i);
    dace(i) = xvace(i)*xbm(i)+zvace(i)*zbm(i)+acediff(i)*aced2(i);
    dsucc(i) = xvsucc(i)*xbm(i)+zvsucc(i)*zbm(i)+succdiff(i)*succd2(i);
    dlac(i) = xvlac(i)*xbm(i)+zvlac(i)*zbm(i)+lacdiff(i)*lacd2(i);
    dpyo(i) = xvpyo(i)*xbm(i)+pyodiff(i)*pyod2(i);
    
    dy = [dy dxbm(i) dzbm(i) dglc(i) do2(i) dace(i) dsucc(i) dlac(i) dpyo(i)];
    
end
% INFO.fluxy = fluxy;

dylast=zeros(1,nmodel);
for i=1:nmodel
    dylast(i) = penalty(i);
end
dy = [dy dylast];
dy = dy';
end
